#include <iomanip.h>
#include <stdlib.h>
#include <fstream.h>
#include <strstream.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include "client.h"

extern "C"
{
#include "aim.h" /* libfaim */
}

char cTempCrud[1024];	//To be unneeded when this is fully C++...

void errUsageGlobal()
{
	cout << "Usage:" << endl;
	cout << endl;
	cout << "bonim --help" << endl;
	cout << "bonim [[-c] [-d] [-p | -P] profile] [--msg target message]" << endl;
	cout << endl;
	cout << "Usage options:" << endl;
	cout << "  profile: The ID of the profile to login as." << endl;
	cout << "           If left out it will use the default profile." << endl;
	cout << endl;
	cout << "  -c:      Create configs for the profile." << endl;
	cout << "           Use this for a screen name that you have not" << endl;
	cout << "           logged in with Bonim before but is already" << endl;
	cout << "           registered with AOL." << endl;
	cout << endl;
	cout << "  -d:      Make the passed profile the default profile." << endl;
	cout << "           It will then be chosen automatically when you" << endl;
	cout << "           launch Bonim but specify no profile in the parameters." << endl;
	cout << endl;
	cout << "  -p:      Do not store the password in the profile." << endl;
	cout << endl;
	cout << "  -P:      Store the password in the profile. This is default" << endl;
	cout << "           when creating a new profile. It can also be used with" << endl;
	cout << "           an existing profile to undo the option to not store" << endl;
	cout << "           your password." << endl;
	cout << endl;
	cout << "  --msg:   This will, upon successful login, send the message you" << endl;
	cout << "           pass to the target user you pass in. Then it will quit" << endl;
	cout << "           It is like a hit-and-run IM, useful for shell scripts." << endl;
	cout << "           Note that if you do this too often in too shorty a time" << endl;
	cout << "           span then the AOL servers will stop letting you log in" << endl;
	cout << "           temporarily. Also not that if you want to do this without" << endl;
	cout << "           seeing the Bonim window then redurect the output to" << endl;
	cout << "           /dev/null by adding \" > /dev/null\" at the end of the command" << endl;
	cout << "           you type to launch Bonim." << endl;
}

void quitBonimInt(int blah)
{
	wprintw(BONIM->mainwin, "This has been a bonehead production.\n\r");
	wrefresh(BONIM->mainwin);
	
	delete BONIM;
	
	exit(0);
}

Client *Client::m_instance=0;

Client::Client()
{
	m_version="1.0.5";
	
	setTarget("<none>");
	setLastFrom("--- None ---");
	setLastTo("--- None ---");
	setRunStatus(crsRun);
	
	mkdir((String(getenv("HOME"))+"/.bonim").toCString(), 0755);
	
	//Start up curses...
	initscr();				/* initialize the curses library */
	keypad(stdscr, TRUE);	/* enable keyboard mapping */
	nl();					/* tell curses to do NL->CR/NL on output */
	cbreak();				/* take input chars one at a time, no wait for \n\r */
	noecho();				/* do not echo input */
	intrflush(stdscr, FALSE);
	
	if(has_colors())
	{
		start_color();
		
		// Standard colours
		init_pair(0, COLOR_WHITE, COLOR_RED);
		init_pair(1, COLOR_RED, COLOR_BLACK);
		init_pair(2, COLOR_GREEN, COLOR_BLACK);
		init_pair(3, COLOR_YELLOW, COLOR_BLACK);
		init_pair(4, COLOR_BLUE, COLOR_BLACK);
		init_pair(5, COLOR_MAGENTA, COLOR_BLACK);
		init_pair(6, COLOR_CYAN, COLOR_BLACK);
		init_pair(7, COLOR_WHITE, COLOR_BLACK);
		
		// Extra colours used in status messages etc.
		init_pair(8, COLOR_WHITE, COLOR_RED);
		init_pair(9, COLOR_WHITE, COLOR_GREEN);
		init_pair(10, COLOR_WHITE, COLOR_BLUE);
		
		init_pair(11, COLOR_RED, COLOR_WHITE);
		init_pair(12, COLOR_GREEN, COLOR_WHITE);
		init_pair(13, COLOR_BLUE, COLOR_WHITE);
		
		// Inverse colours
		init_pair(16, COLOR_BLACK, COLOR_WHITE);
		init_pair(17, COLOR_RED, COLOR_WHITE);
		init_pair(18, COLOR_GREEN, COLOR_WHITE);
		init_pair(19, COLOR_YELLOW, COLOR_WHITE);
		init_pair(20, COLOR_BLUE, COLOR_WHITE);
		init_pair(21, COLOR_MAGENTA, COLOR_WHITE);
		init_pair(22, COLOR_CYAN, COLOR_WHITE);
		init_pair(23, COLOR_WHITE, COLOR_RED);
	}
	
	wattron(stdscr, COLOR_PAIR(5));
	
	if(COLS<91) m_screenSize=sszSmall;
	else if(COLS > 90 && COLS < 121) m_screenSize=sszMedium;
	else m_screenSize=sszLarge;
	
	setLogIMs(false);
	m_visStatus=visAll;
	setBeepOnMsg(true);
	setShowOffline(false);
	setShowAlias(false);
	setShowSigns(true);
	
	m_storePassFlagSet=false;
}

Client::~Client()
{
	endwin();
}

Client *Client::gClient()
{
	return m_instance?m_instance:m_instance=new Client;
}
const String &Client::getVersion()
{
	return m_version;
}

bool Client::addBud(const String &budID)
{
	Buddy *newBud;
	
	newBud=new Buddy;
	
	if(getBudByID(budID).budID!="" || !newBud) return false;
	
	newBud->budID=budID;
	newBud->budAlias=budID;
	newBud->budStatus=avsOffLine;
	newBud->budTime="--:--";
	newBud->budClass=uclUnknown;
	
	if(!m_buds.addItem(newBud, m_buds.getSize()))
	{
		delete newBud;
		
		return false;
	}
	
	return true;
}

bool Client::setBudAlias(const String &budID, const String &budAlias)
{
	Buddy *updateMe;
	
	updateMe=getBuddyPtrByID(budID);
	
	if(!updateMe) return false;
	
	updateMe->budAlias=budAlias;
	
	return true;
}

bool Client::setBudClass(const String &budID, UserClass budClass)
{
	Buddy *updateMe;
	
	updateMe=getBuddyPtrByID(budID);
	
	if(!updateMe) return false;
	
	updateMe->budClass=budClass;
	
	return true;
}

bool Client::setBudStatus(const String &budID, AvailStatus budStatus)
{
	Buddy *updateMe;
	
	updateMe=getBuddyPtrByID(budID);
	
	if(!updateMe) return false;
	
	updateMe->budStatus=budStatus;
	
	return true;
}

bool Client::setBudTime(const String &budID, String &budTime)
{
	Buddy *updateMe;
	
	updateMe=getBuddyPtrByID(budID);
	
	if(!updateMe) return false;
	
	updateMe->budTime=budTime;
	
	return true;
}

unsigned long Client::getBudCount()
{
	return m_buds.getSize();
}

Buddy Client::getBudByPosition(unsigned long budPosition)
{
	Buddy corpseOfABuddy;
	Buddy *returnMe;
	
	returnMe=m_buds.getItem(budPosition);
	
	if(!returnMe) return corpseOfABuddy;
	
	return *returnMe;
}

Buddy Client::getBudByID(const String &budID)
{
	Buddy corpseOfABuddy;
	Buddy *returnMe;
	
	returnMe=getBuddyPtrByID(budID);
	
	if(!returnMe) return corpseOfABuddy;
	
	return *returnMe;
}

Buddy Client::getBudByAlias(String budAlias)
{
	unsigned int iter;
	Buddy *iterBud;
	Buddy corpseOfABuddy;
	
	budAlias=unformatBudID(budAlias);
	
	for(iter=0, iterBud=m_buds.getItem(0); iter<m_buds.getSize(); iter++, iterBud=m_buds.getItem(iter))
	{
		if(!iterBud) return corpseOfABuddy;
		
		if(budAlias==unformatBudID(iterBud->budAlias)) return *iterBud;
	}
	
	return corpseOfABuddy;
}

bool Client::killBud(String budID)
{
	unsigned int iter;
	Buddy *iterBud;
	
	budID=unformatBudID(budID);
	
	for(iter=0, iterBud=m_buds.getItem(0); iter<m_buds.getSize(); iter++, iterBud=m_buds.getItem(iter))
	{
		if(!iterBud) return false;
		
		if(budID==unformatBudID(iterBud->budID)) return m_buds.killItem(iter);
	}
	
	return false;
}

void Client::killBuds()
{
	m_buds.kill();
}

void Client::chopInputLine()
{
	m_inputLine=m_inputLine.left(m_inputLine.getLength()-1);
}

void Client::clearInputLine()
{
	m_inputLine="";
}

void Client::appendInputLine(char newChar)
{
	m_inputLine+=newChar;
}

const String &Client::getInputLine() const
{
	return m_inputLine;
}

bool Client::setUserID(const String &newUserID)
{
	if(m_userID!="") return false;
	
	m_userID=newUserID;
	
	return true;
}

const String &Client::getUserID() const
{
	return m_userID;
}

void Client::setPassword(const String &newPassword)
{
	m_password=newPassword;
}

const String &Client::getPassword() const
{
	return m_password;
}

void Client::setTarget(const String &newTarget)
{
	m_target=newTarget;
}

const String &Client::getTarget() const
{
	return m_target;
}

String Client::getStatus() const
{
	if(getAwayMsg()=="") return "Online (avail)";
	
	return "Online (away)";
}

String Client::getBonimDir() const
{
	return String(getenv("HOME"))+"/.bonim/";
}

String Client::getConfigDir() const
{
	return String(getenv("HOME"))+"/.bonim/"+getUserID()+'/';
}

void Client::setAwayMsg(const String &newAwayMsg)
{
	m_awayMsg=newAwayMsg;
}

const String &Client::getAwayMsg() const
{
	return m_awayMsg;
}

void Client::unaway()
{
	setAwayMsg("");
	
	wattron(mainwin, COLOR_PAIR(1));
	wprintw(mainwin, "Away status removed.\n\r");
	wrefresh(mainwin);
	wattron(mainwin, COLOR_PAIR(6));
}

void Client::setLogIMs(bool newLogIMs)
{
	m_logIMs=newLogIMs;
}

bool Client::logIMs() const
{
	return m_logIMs;
}

void Client::setVisStatus(VisStatus newVisStatus)
{
	m_visStatus=newVisStatus;
	
	if(newVisStatus==visAll) aim_bos_changevisibility(NULL, AIM_VISIBILITYCHANGE_DENYADD, getUserID().toCString());
	else
	{
		makeBudsOnlyVis(NULL);
	}
}

VisStatus Client::getVisStatus()
{
	return m_visStatus;
}

void Client::setBeepOnMsg(bool newBeepOnMsg)
{
	m_beepOnMsg=newBeepOnMsg;
}

bool Client::beepOnMsg()
{
	return m_beepOnMsg;
}

void Client::setShowOffline(bool newShowOffline)
{
	m_showOffline=newShowOffline;
}

bool Client::showOffline()
{
	return m_showOffline;
}

void Client::setShowAlias(bool newShowAlias)
{
	m_showAlias=newShowAlias;
}

bool Client::showAlias()
{
	return m_showAlias;
}

void Client::setShowSigns(bool newShowSigns)
{
	m_showSigns=newShowSigns;
}

bool Client::showSigns()
{
	return m_showSigns;
}	

void Client::setLowColDisp(LowColDisp newLowColDisp)
{
	m_lowColDisp=newLowColDisp;
}

LowColDisp Client::getLowColDisp()
{
	return m_lowColDisp;
}

void Client::setLastFrom(const String &newLastFrom)
{
	m_lastFrom=newLastFrom;
}

String &Client::getLastFrom()
{
	return m_lastFrom;
}

void Client::setLastTo(const String &newLastTo)
{
	m_lastTo=newLastTo;
}

String &Client::getLastTo()
{
	return m_lastTo;
}

ScreenSize Client::getScreenSize()
{
	return m_screenSize;
}

void Client::setRunStatus(ClientRunStatus newRunStatus)
{
	m_runStatus=newRunStatus;
}

ClientRunStatus Client::getRunStatus() const
{
	return m_runStatus;
}

void Client::errUsage()
{
	endwin();
	
	errUsageGlobal();
	
	delete this;	//Heh heh...
	
	exit(1);
}

void Client::showPrompt()
{
	wattron(inputwin, COLOR_PAIR(2));
	wprintw(inputwin,"[%s] ", getTarget().toCString());
	wattron(inputwin, COLOR_PAIR(7));
	wrefresh(inputwin);
}

void Client::displayStatus()
{
	char tid[50];
	int xx;
	
	wclear(statwin);
	
	for (xx=0; xx<6; xx++)
	{
		waddstr(statwin,"\n\r");
	}
	wmove(statwin,0,0);
	
	if(getScreenSize()==sszLarge)
	{
		wattron(statwin, COLOR_PAIR(10));
		wprintw(statwin, "   Bonim AIM for the Linux console.   \n\r");
		strcpy(tid, getUserID().toCString());
		tid[20]='\0';
		wattron(statwin, COLOR_PAIR(6));
		wprintw(statwin, "Logged on as      : ");
		wattron(statwin, COLOR_PAIR(2));
		wprintw(statwin, "%-20s\n\r", tid);
		strcpy(tid, getStatus().toCString());
		tid[20]='\0';
		wattron(statwin, COLOR_PAIR(6));
		wprintw(statwin, "Current status is : ");
		wattron(statwin, COLOR_PAIR(2));
		wprintw(statwin, "%s.\n\r", tid);
		wattron(statwin, COLOR_PAIR(6));
		wprintw(statwin, "Last message from : ");
		wattron(statwin, COLOR_PAIR(2));
		wprintw(statwin, "%-20s\n\r", getLastFrom().toCString());
		strcpy(tid, getLastTo().toCString());
		tid[20]='\0';
		wattron(statwin, COLOR_PAIR(6));
		wprintw(statwin, "Last message to   : ");
		wattron(statwin, COLOR_PAIR(2));
		wprintw(statwin, "%-20s",tid);
	}
	
	else if(getScreenSize()==sszMedium)
	{
		wattron(statwin, COLOR_PAIR(10));
		wprintw(statwin, " Bonim AIM for the console. \n\r");
		strcpy(tid, getUserID().toCString());
		tid[15]='\0';
		wattron(statwin, COLOR_PAIR(6));
		wprintw(statwin, "Logged on as : ");
		wattron(statwin, COLOR_PAIR(2));
		wprintw(statwin, "%-15s\n\r", tid);
		strcpy(tid, getStatus().toCString());
		tid[15]='\0';
		wattron(statwin, COLOR_PAIR(6));
		wprintw(statwin, "Status is    : ");
		wattron(statwin, COLOR_PAIR(2));
		wprintw(statwin, "%-15s\n\r", tid);
		wattron(statwin, COLOR_PAIR(6));
		wprintw(statwin, "Last from    : ");
		wattron(statwin, COLOR_PAIR(2));
		wprintw(statwin, "%-15s\n\r", getLastFrom().toCString());
		strcpy(tid, getLastTo().toCString());
		tid[15]='\0';
		wattron(statwin, COLOR_PAIR(6));
		wprintw(statwin, "Last to      : ");
		wattron(statwin, COLOR_PAIR(2));
		wprintw(statwin, "%-15s", tid);
	}
	else
	{
		strcpy(tid, getUserID().toCString());
		tid[15]='\0';
		wattron(statwin, COLOR_PAIR(6));
		wprintw(statwin, "Nick: ");
		wattron(statwin, COLOR_PAIR(2));
		wprintw(statwin, "%-15s\n\r", tid);
		strcpy(tid, getStatus().toCString());
		tid[15]='\0';
		wattron(statwin, COLOR_PAIR(6));
		wprintw(statwin, "Stat: ");
		wattron(statwin, COLOR_PAIR(2));
		wprintw(statwin, "%-15s\n\r", tid);
		wattron(statwin, COLOR_PAIR(6));
		wprintw(statwin, "L fr: ");
		wattron(statwin, COLOR_PAIR(2));
		wprintw(statwin, "%-15s\n\r", getLastFrom().toCString());
		strcpy(tid, getLastTo().toCString());
		tid[15]='\0';
		wattron(statwin, COLOR_PAIR(6));
		wprintw(statwin, "L to: ");
		wattron(statwin, COLOR_PAIR(2));
		wprintw(statwin, "%-15s",tid);
	}
	
	wrefresh(statwin);
}

void Client::showUserList()
{
	unsigned long iter;
	unsigned long onLineCount;
	String budItem;
	Buddy iterBud;
	
	wclear(listwin);
	
	wattron(listwin, COLOR_PAIR(1));
	
	for(iter=0, iterBud=getBudByPosition(0); iter<getBudCount(); iter++, iterBud=getBudByPosition(iter))
	{
		if(iterBud.budStatus==avsOffLine && showOffline())
		{
			if(getScreenSize()==sszLarge)
			{
				if(showAlias()) budItem="   "+iterBud.budAlias;
				else budItem="   "+iterBud.budID;
				
				if(budItem.getLength()>18) budItem=budItem.left(18);
				else
				{
					while(budItem.getLength()<18) budItem+=' ';
				}
				
				budItem+="[Offline] "+iterBud.budTime;
				
				if(iterBud.budClass==uclTrial) budItem+=" [T]\n\r";
				else if(iterBud.budClass==uclAOL) budItem+=" [A]\n\r";
				else if(iterBud.budClass==uclFree) budItem+=" [F]\n\r";
				else if(iterBud.budClass==uclUnknown) budItem+=" [?]\n\r";
			}
			else if(getScreenSize()==sszMedium)
			{
				if(showAlias()) budItem=iterBud.budAlias;
				else budItem=iterBud.budID;
				
				if(budItem.getLength()>18) budItem=budItem.left(18);
				else
				{
					while(budItem.getLength()<18) budItem+=' ';
				}
				
				budItem+="[Off] "+iterBud.budTime+"\n\r";
			}
			else
			{
				if(showAlias()) budItem=iterBud.budAlias;
				else budItem=iterBud.budID;
				
				if(budItem.getLength()>17) budItem=budItem.left(17);
				else
				{
					while(budItem.getLength()<17) budItem+=' ';
				}
				
				if(getLowColDisp()==lcdStatus) budItem+="[Off]";
				else budItem+=iterBud.budTime;
			}
			
			waddstr(listwin, budItem.toCString());
		}
		
	}
	
	wattron(listwin, COLOR_PAIR(6));
	
	for(iter=0, iterBud=getBudByPosition(iter); iter<getBudCount(); iter++, iterBud=getBudByPosition(iter))
	{
		if(iterBud.budStatus==avsNew && showOffline())
		{
			if(getScreenSize()==sszLarge)
			{
				if(showAlias()) budItem="   "+iterBud.budAlias;
				else budItem="   "+iterBud.budID;
				
				if(budItem.getLength()>18) budItem=budItem.left(18);
				else
				{
					while(budItem.getLength()<18) budItem+=' ';
				}
				
				budItem+="[New]     "+iterBud.budTime;
				
				if(iterBud.budClass==uclTrial) budItem+=" [T]\n\r";
				else if(iterBud.budClass==uclAOL) budItem+=" [A]\n\r";
				else if(iterBud.budClass==uclFree) budItem+=" [F]\n\r";
				else if(iterBud.budClass==uclUnknown) budItem+=" [?]\n\r";
			}
			else if(getScreenSize()==sszMedium)
			{
				if(showAlias()) budItem=iterBud.budAlias;
				else budItem=iterBud.budID;
				
				if(budItem.getLength()>18) budItem=budItem.left(18);
				else
				{
					while(budItem.getLength()<18) budItem+=' ';
				}
				
				budItem+="[New] "+iterBud.budTime+"\n\r";
			}
			else
			{
				if(showAlias()) budItem=iterBud.budAlias;
				else budItem=iterBud.budID;
				
				if(budItem.getLength()>17) budItem=budItem.left(17);
				else
				{
					while(budItem.getLength()<17) budItem+=' ';
				}
				
				if(getLowColDisp()==lcdStatus) budItem+="[New]";
				else budItem+=iterBud.budTime;
			}
			
			waddstr(listwin, budItem.toCString());
		}
	}
	
	wattron(listwin, COLOR_PAIR(2));
	onLineCount=0;
	
	for(iter=0, iterBud=getBudByPosition(iter); iter<getBudCount(); iter++, iterBud=getBudByPosition(iter))
	{
		if(iterBud.budStatus==avsOnLine)
		{
			if(getScreenSize()==sszLarge)
			{
				onLineCount++;
				
				if(onLineCount<10) budItem=String(onLineCount)+"  ";
				else if(onLineCount<100) budItem=String(onLineCount)+" ";
				else budItem=onLineCount;
				
				if(showAlias()) budItem+=iterBud.budAlias;
				else budItem+=iterBud.budID;
				
				
				if(budItem.getLength()>18) budItem=budItem.left(18);
				else
				{
					while(budItem.getLength()<18) budItem+=' ';
				}
				
				budItem+="[Online]  "+iterBud.budTime;
				
				if(iterBud.budClass==uclTrial) budItem+=" [T]\n\r";
				else if(iterBud.budClass==uclAOL) budItem+=" [A]\n\r";
				else if(iterBud.budClass==uclFree) budItem+=" [F]\n\r";
				else if(iterBud.budClass==uclUnknown) budItem+=" [?]\n\r";
			}
			else if(getScreenSize()==sszMedium)
			{
				if(showAlias()) budItem=iterBud.budAlias;
				else budItem=iterBud.budID;
				
				if(budItem.getLength()>18) budItem=budItem.left(18);
				else
				{
					while(budItem.getLength()<18) budItem+=' ';
				}
				
				budItem+="[On]  "+iterBud.budTime+"\n\r";
			}
			else
			{
				if(showAlias()) budItem=iterBud.budAlias;
				else budItem=iterBud.budID;
				
				if(budItem.getLength()>17) budItem=budItem.left(17);
				else
				{
					while(budItem.getLength()<17) budItem+=' ';
				}
				
				if(getLowColDisp()==lcdStatus) budItem+="[On] ";
				else budItem+=iterBud.budTime;
			}
			
			waddstr(listwin, budItem.toCString());
		}
	}
	
	wattron(listwin, COLOR_PAIR(7));
	wrefresh(listwin);
}

void Client::showNames()
{
	char id[20];
	char tid[20];
	unsigned long iter;
	int cc=0;
	Buddy iterBud;
	
	wattron(mainwin, COLOR_PAIR(6));
	wprintw(mainwin, "Displaying full userlist, %d users found\n\r", getBudCount());
	
	for(iter=0, iterBud=getBudByPosition(0); iter<getBudCount(); iter++, iterBud=getBudByPosition(iter))
	{
		strcpy(id, iterBud.budID.toCString());
		
		if(iterBud.budStatus==avsOffLine) wattron(mainwin, COLOR_PAIR(1));
		else if(iterBud.budStatus==avsOnLine) wattron(mainwin, COLOR_PAIR(2));
		else if(iterBud.budStatus==avsNew) wattron(mainwin, COLOR_PAIR(3));
//		else if(iterBud.budStatus=='3') wattron(mainwin, COLOR_PAIR(4));
		
		cc=cc+19;
		
		if (cc >= maincols)
		{
			wprintw(mainwin,"\n");
			cc=19;
		}
		
/*		for(xx = 0; xx < users; xx++)
		{
			if(strcasecmp(id, iterBud.budID.toCString())==0)
			un=xx;
		}
		
		strcpy(tid,userarray[un][1]);*/
		
		strcpy(tid, iterBud.budID.toCString());
		
		if(showAlias())
		{
			strcpy(tid, iterBud.budAlias.toCString());
			tid[MAXSNLEN]='\0';
		}
		wprintw(mainwin, "[%-16s] ",tid);
	}
	wattron(mainwin, COLOR_PAIR(7));
	wprintw(mainwin,"\n");
	wrefresh(mainwin);
	
	return;
}

void Client::clearAll()
{
	wclear(mainwin);
	wrefresh(mainwin);
	wclear(inputwin);
	wrefresh(inputwin);
	showUserList();
}

void Client::displayHelp(void)
{
	system("bonim --help | less");
	
	clearAll();
}

void Client::readConfig()
{
	ifstream fIn;
	int configVersion;
	String inputLine;
	String budID;
	char configlocation[256];
	unsigned long iter;
	unsigned long numBuds;
	
	if(chdir(getConfigDir().toCString()))
	{
		endwin();
		
		cout << "Could not find directory \"" << getConfigDir() << "\"." << endl;
		
		errUsage();
	}
	
	strcpy(configlocation, getConfigDir().toCString());
	strcat(configlocation, "config");
	
	killBuds();
	
	fIn.open(configlocation);
	
	if(!fIn)
	{
		endwin();
		
		printf("Could not open \"%s\".\r\n\r\n", configlocation);
		
		errUsage();
	}
	
	configVersion=inputLine.getLine(fIn, 1024).toShort();
	
	setPassword(inputLine.getLine(fIn, 1024));
	
	if(getPassword()=='!' && !m_storePassFlagSet) m_storePass=false;
	else if(getPassword()=='?' && !m_storePassFlagSet) m_storePass=true;
	
	setBeepOnMsg(configVersion==1?inputLine.getLine(fIn, 1024).toShort():inputLine.getLine(fIn, 1024).toBool());
	setShowAlias(configVersion==1?inputLine.getLine(fIn, 1024).toShort():inputLine.getLine(fIn, 1024).toBool());
	setLowColDisp(LowColDisp(inputLine.getLine(fIn, 1024).toShort()));
	
	numBuds=inputLine.getLine(fIn, 1024).toShort();
	
	for(iter=0; iter<numBuds; iter++)
	{
		addBud(budID=inputLine.getLine(fIn, 1024));
	
		AOLUserList+='&'+inputLine;
		
		setBudAlias(budID, inputLine.getLine(fIn, 1024));	//Will this even work? Depends on the order of evaluation...
	}
	
	setShowOffline(configVersion==1?inputLine.getLine(fIn, 1024).toShort():inputLine.getLine(fIn, 1024).toBool());
	setLogIMs(configVersion==1?inputLine.getLine(fIn, 1024).toShort():inputLine.getLine(fIn, 1024).toBool());
	//Put this line back and remove the one after it in when we get better integration with libfaim...
	//setVisStatus(VisStatus(inputLine.getLine(fIn, 1024).toShort()));
	m_visStatus=VisStatus(inputLine.getLine(fIn, 1024).toShort());
	setShowSigns(configVersion==1?true:inputLine.getLine(fIn, 1024).toBool());
	
	fIn.close();
	
	wrefresh(listwin);
	showUserList();
}

void Client::writeConfig()
{
	ofstream fOut;
	char configlocation[256];
	unsigned long iter;
	Buddy iterBud;
	
	strcpy(configlocation, getConfigDir().toCString());
	strcat(configlocation, "config");
	
	fOut.open(configlocation);
	
	fOut << "2\n";
	fOut << (m_storePass?getPassword():'!') << '\n';
	fOut << String(beepOnMsg()) << '\n';
	fOut << String(showAlias()) << '\n';
	fOut << String(short(getLowColDisp())) << '\n';
	fOut << String(getBudCount()) << '\n';
	
	for(iter=0, iterBud=getBudByPosition(0); iter<getBudCount(); iter++, iterBud=getBudByPosition(iter))
	{
		fOut << iterBud.budID << '\n';
		fOut << iterBud.budAlias << '\n';
	}
	
	fOut << String(showOffline()) << '\n';
	fOut << String(logIMs()) << '\n';
	fOut << String(short(getVisStatus())) << '\n';
	fOut << String(showSigns()) << '\n';
	fOut << "***\n";
	
	fOut.close();
}

void Client::makeConfig()
{
	char configlocation[256];
	ofstream fOut;
	
	strcpy(configlocation, getConfigDir().toCString());
	strcat(configlocation, "config");
	
	fOut.open(configlocation);
	
	if(!fOut)
	{
		printf("Could not create \"%s\".\r\n\r\n", configlocation);
		
		errUsage();
	}
	
	fOut << "2\n";
	fOut << (m_storePass?"?\n":"!\n");
	fOut << "true\n";
	fOut << "false\n";
	fOut << short(lcdStatus) << "\n";
	fOut << "1\n";
	fOut << "The Extern Void\n";
	fOut << "The Extern Void\n";
	fOut << "1\n";
	fOut << "false\n";
	fOut << short(visAll) << "\n";
	fOut << "true\n";
	fOut << "***\n";
	
	fOut.close();
	
	wattron(mainwin, COLOR_PAIR(1));
	wprintw(mainwin, "\n\rFile created: %s\n\r", configlocation);
	wattron(mainwin, COLOR_PAIR(6));
	wrefresh(mainwin);
}

void Client::makeBudsOnlyVis(struct aim_conn_t *conn)
{
	char visList[4096];
	unsigned long iter;
	Buddy iterBud;
	
	visList[0]='\0';
	
	for(iter=0, iterBud=getBudByPosition(0); iter<getBudCount(); iter++, iterBud=getBudByPosition(iter))
	{
		strcat(visList, iterBud.budID.toCString());
		strcat(visList, "&");
	}
	
	aim_bos_changevisibility(conn, AIM_VISIBILITYCHANGE_PERMITADD, visList);
}

int Client::getBuddy(char *out, const char *in, int searchstatus, bool onlinestatus)
//Based on user input "in", find a user and store it in "out".
//Looks for a match on your buddy list, aliases, etc. If none found
//then it returns the same name you passed it. If you only want it to
//find a user ID on the buddy list (not alias list or anything else) then
//pass 2 for the variable searchstatus. Pass it 1 to search buddy and alias
//list. Pass it 0 to search both lists and to search with in as a number
//for the "in"-th person. Pass 0 for onlinestatus to ignore on-line status;
//pass it 1 to match only on-line people. It returns -1 if it failed to
//make a match. When this happens, in is copied to out. If it finds a match,
//it returns its index in the buddy list.
{
	unsigned long iter;
	int foundCount;
	int countGoal;
	String userID;
	Buddy iterBud;
	
	userID=String(in).flattenTokens().toLower();
	
	//Check for match in buddy list. We do this because we want to match
	//the input "deusdec" to buddy "Deus de C". In otherwords, we search
	//without case nor white space but return the UID that has the space
	//and case formating.
	for(iter=0, iterBud=getBudByPosition(0); iter<getBudCount(); iter++, iterBud=getBudByPosition(iter))
	{
		if(userID==iterBud.budID.flattenTokens().toLower() && (iterBud.budStatus==avsOnLine || !onlinestatus))
		{

			strcpy(out, iterBud.budID.toCString());
			
			return iter;
		}
	}
	
	if(searchstatus==2) return -1;
	
	//Ditto alisas
	for(iter=0, iterBud=getBudByPosition(0); iter<getBudCount(); iter++, iterBud=getBudByPosition(iter))
	{
		if(userID==iterBud.budAlias.flattenTokens().toLower() && (iterBud.budStatus==avsOnLine || !onlinestatus))
		{
			strcpy(out, iterBud.budID.toCString());
			
			return iter;
		}
	}
	
	if(searchstatus) return -1;
	
	userID=String(in).stripTrailing().stripTrailing();
	
	if(!userID.isNumber()) return -1;
	
	if(float(userID.toLong())!=userID.toFloat()) return -1;
	
	countGoal=userID.toLong();
	foundCount=0;
	
	for(iter=0, iterBud=getBudByPosition(0); iter<getBudCount(); iter++, iterBud=getBudByPosition(iter))
	{
		if(iterBud.budStatus==avsOnLine || !onlinestatus) foundCount++;
		
		if(foundCount==countGoal)
		{
			strcpy(out, iterBud.budID.toCString());
			
			return iter;
		}
	}
	cout << "Fail" << endl;
	strcpy(out, in);
	
	return -1;
}

void Client::addUser()
{
	char tnick[20];
	strcpy(tnick, &getInputLine().toCString()[5]);
	wprintw(mainwin, "Adding user [%s], stand by...\n\r",tnick);
	
	if(!addBud(tnick))
	{
		wprintw(mainwin, "There was an error, probably due to low memory(!)\n\r");
		
		return;
	}
	writeConfig();
	wprintw(mainwin, "Done, you will need to restart the client for\n\rthese changes to take full effect.\n\r");
	showUserList();
}

void Client::delUser()
{
	wprintw(mainwin, "Removing user; stand by...\n\r");
	
	if(!killBud(getInputLine().getTokens(1, 0))) wprintw(mainwin, "Unable to remove user, probaly because the user is not in list.\n\r");
	
	writeConfig();
	showUserList();
}

long Client::go(int args, char *argv[])
{
	char defprofloc[256];
	
	int iter;
	int iter2;
	
	fd_set rfds;
	struct timeval tv;
	int retval;
	
#if 0
	/*
	specify your custom command handlers here.  I recommend
	at least overriding the login_phase3d_f handler so that
	you can have your own buddy list and profile.  The
	rest are probably only useful to override for UI
	reasons.
	*/
	rxcallback_t faimtest_callbacks[] = {
		faimtest_parse_incoming_im, /* incoming IM 0 */
		faimtest_parse_oncoming, /* oncoming buddy 1 */
		faimtest_parse_offgoing, /* offgoing buddy 2 */
		faimtest_parse_misses, /* AIM_CB_MISSED_IM 3 */
		faimtest_parse_misses, /* AIM_CB_MISSED_CALL 4 */
		NULL, /* login phase 4 packet C command 1 -- depricated 5 */
		NULL, /* login phase 4 packet C command 2 -- depricated 6 */
		NULL, /* login phase 2, first resp -- depricated 7 */
		faimtest_serverready, /* server ready -- **HANDLING REQUIRED** 8 */
		NULL, /* login phase 3 packet B -- depricated 9 */
		NULL, /* login phase 3D packet A -- depricated 10 */
		NULL, /* login phase 3D packet B -- depricated 11 */
		NULL, /* login phase 3D packet C -- depricated 12 */
		NULL, /* login phase 3D packet D -- depricated 13 */
		NULL, /* login phase 3D packet E -- depricated 14 */
		faimtest_handleredirect, /* redirect -- **HANDLING REQUIRED** 15 */
		faimtest_parse_misses, /* AIM_CB_RATECHANGE 16 */
		faimtest_parse_misses, /* AIM_CB_USERERROR 17 */
		NULL, /* completely unknown command 18 */
		faimtest_parse_userinfo, /* user info response 19 */
		NULL, /* user search by address response 20 */
		NULL, /* user serach by name response 21 */
		NULL, /* user search fail 22 */
		faimtest_auth_error, /* 23 */
		faimtest_auth_success, /* 24 */
		faimtest_authsvrready, /* 25 */
		NULL, /* 26 */
		faimtest_pwdchngdone, /* 27 */
		faimtest_serverready, /* 28 */
		0x00 /* terminating NULL -- REQUIRED 29 */
	};
#endif
	struct client_info_s info = {/*("Bonim "+getVersion()+" (Phear!)").toCString()*/"Bonim 1.0.5 (Phear!)", 3, 90, 42, "us", "en"};
	struct aim_conn_t *authconn = NULL;
	
	//Vars used in processing command line args.
	int initaction;  //0==open, 1==create new
	int initsetdefault;
	bool hackArgMustHaveUserID;
	
	/* Break signals can screw up your terminal so let's stop those. */
	struct sigaction ignoreMe;
	
	ignoreMe.sa_handler=quitBonimInt;
	sigaction(SIGINT, &ignoreMe, NULL);
	
	aim_connrst(); /* reset connection array -- there's a better place for this*/
	/* register our callback array (optional) */
	//aim_register_callbacks(faimtest_callbacks);
	
	strcpy(defprofloc, getBonimDir().toCString());
	strcat(defprofloc, "defaultprofile");
	
	initaction=0;
	m_storePass=true;
	initsetdefault=0;
	m_storePassFlagSet=false;
	
	hackArgMustHaveUserID=false;
	
	for(iter=1; iter<args; iter++)
	{
		if(String(argv[iter])=="--msg")
		{
			if(m_hitAndRunTarget!="") errUsage();
			
			iter++;
			if(iter==args) errUsage();
			
			iter++;
			if(iter==args) errUsage();
			
			m_hitAndRunTarget=String(argv[iter-1]).stripLeading().stripTrailing();
			m_hitAndRunMessage=String(argv[iter]).stripLeading().stripTrailing();
			
			if(m_hitAndRunTarget=="" || m_hitAndRunMessage=="") errUsage();
		}
		else if(argv[iter][0]=='-')
		{
			hackArgMustHaveUserID=true;
			
			for(iter2=1; argv[iter][iter2]!='\0'; iter2++)
			{
				if(argv[iter][iter2]=='c')
				{
					if(initaction) errUsage();
					
					initaction=1;
				}
				else if(argv[iter][iter2]=='d')
				{
					if(initsetdefault) errUsage();
					
					initsetdefault=true;
				}
				else if(argv[iter][iter2]=='p')
				{
					if(m_storePassFlagSet) errUsage();
					
					m_storePassFlagSet=true;
					m_storePass=false;
				}
				else if(argv[iter][iter2]=='P')
				{
					if(m_storePassFlagSet) errUsage();
					
					m_storePassFlagSet=true;
					m_storePass=true;
				}
				else errUsage();
			}
		}
		else if(!setUserID(argv[iter])) errUsage();
	}
	
	if(getUserID()=="" && hackArgMustHaveUserID) errUsage();
	
	if(getUserID()=="")
	{
		String inputLine;
		ifstream fIn(defprofloc);
		
		if(!fIn)
		{
			endwin();
			printf("\n\nFatal! %s not found!\n", defprofloc);
			errUsage();
			
			return 1;
		}
		
		inputLine.getLine(fIn, 1024);
		
		if(inputLine=="")
		{
			printf("\n\nFatal! %s could not be read or is empty.\n", defprofloc);
			
			return 1;
		}
		
		fIn.close();
		
		setUserID(inputLine);
	}
	
	if(initaction)
	{
		if(mkdir(getConfigDir().toCString(), 0755))
		{
			cout << "Could not create directory \"" << getConfigDir() << "\".\n\rMaybe it already exists?\n\r" << endl;
			
			errUsage();
		}
		
		mkdir((getConfigDir()+"logs").toCString(), 0755);
	}
	
	if(initsetdefault)
	{
		ofstream fOut;
		
		fOut.open(defprofloc);
		
		if(!fOut)
		{
			printf("Could not write to \"%s\".\n\r\n\r", defprofloc);
			
			errUsage();
		}
		
		fOut << getUserID();
		
		fOut.close();
	}
	
	if(getScreenSize()==sszSmall)
	{
		inputwin = subwin(stdscr, 4, COLS-25, LINES-5, 1);
		move(LINES-6,1);
		hline(0,COLS-2);
		mainwin = subwin(stdscr, LINES-7,COLS-25,1,1);
		move(1,COLS-24);
		vline(0,LINES-1);
		listwin = subwin(stdscr, LINES-7,22,1,COLS-23);
		statwin = subwin(stdscr, 4, 22,LINES-5,COLS-23);
		box(stdscr,0,0);
		maincols = (COLS-25);
	}
	else if(getScreenSize()==sszMedium)
	{
		inputwin = subwin(stdscr, 5, COLS-34, LINES-6, 1);
		move(LINES-7,1);
		hline(0,COLS-2);
		mainwin = subwin(stdscr, LINES-8,COLS-34,1,1);
		move(1,COLS-33);
		vline(0,LINES-1);
		listwin = subwin(stdscr, LINES-8,31,1,COLS-32);
		statwin = subwin(stdscr, 5, 31,LINES-6,COLS-32);
		box(stdscr,0,0);
		maincols = (COLS-34);
	}
	else
	{
		inputwin = subwin(stdscr, 5, COLS-44, LINES-6, 1);
		move(LINES-7,1);
		hline(0,COLS-2);
		mainwin = subwin(stdscr, LINES-8,COLS-44,1,1);
		move(1,COLS-43);
		vline(0,LINES-1);
		listwin = subwin(stdscr, LINES-8,41,1,COLS-42);
		statwin = subwin(stdscr, 5, 41,LINES-6,COLS-42);
		box(stdscr,0,0);
		maincols = (COLS-44);
	}
	
	scrollok(inputwin, TRUE);
	idlok(inputwin, TRUE);
	scrollok(mainwin, TRUE);
	idlok(mainwin, TRUE);
	scrollok(listwin, TRUE);
	idlok(listwin, TRUE);
	scrollok(statwin, TRUE);
	idlok(statwin, TRUE);
	
	wattron(stdscr, COLOR_PAIR(6));
	wattron(inputwin, COLOR_PAIR(6));
	wattron(mainwin, COLOR_PAIR(6));
	wattron(listwin, COLOR_PAIR(6));
	wattron(statwin, COLOR_PAIR(6));
	
	wattron(mainwin, COLOR_PAIR(2));
	wprintw(mainwin, (char *)("Bonim AIM client "+getVersion()+" for the Linux console.\n\r").toCString());
	wprintw(mainwin, "(c) 2001 BoneHead Productions (David Mooter).\n\r");
	wprintw(mainwin, "Your display mode seems to be %dx%d\n\r", COLS, LINES);
	wattron(mainwin, COLOR_PAIR(7));
	wprintw(mainwin, "Type /HELP for help on commands.\n\r");
	wattron(mainwin, COLOR_PAIR(6));
	
	refresh();
	
	if(initaction) makeConfig();
	
	readConfig();
	
	wprintw(mainwin, "\n\rPlease enter your password... ");
	
	wprintw(inputwin, "Username: %s\n\r", getUserID().toCString());
	wprintw(inputwin, "Password: ");
	wrefresh(mainwin);
	wrefresh(inputwin);
	if(getPassword()=='?' || getPassword()=='!')
	{
		scanf("%s", cTempCrud);
		setPassword(cTempCrud);
	}
	wprintw(inputwin, "<hidden>\n\r");
	wrefresh(inputwin);
	wprintw(mainwin, "Thank you.\n\rConnecting to server, please stand by...\n\r\n\r");
	wrefresh(mainwin);
	
	while(getRunStatus()!=crsQuit)
	{
		showUserList();
		displayStatus();
		showPrompt();
		
		setRunStatus(crsRun);	//Reset the run status in case we just looped with crsRestart run status
		
		authconn = aim_newconn(AIM_CONN_TYPE_AUTH, FAIM_LOGIN_SERVER);
		
		if (authconn == NULL)
		{
			wprintw(mainwin, "BONIM: internal connection error while in aim_login.  bailing out.\n\r");
			wrefresh(mainwin);
			return -1;
		}
		else if (authconn->fd == -1)
		{

			if (authconn->status & AIM_CONN_STATUS_RESOLVERR) wprintw(mainwin, "BONIM: could not resolve authorizer name\n\r");
			else if (authconn->status & AIM_CONN_STATUS_CONNERR) wprintw(mainwin, "BONIM: could not connect to authorizer\n\r");
			
			wrefresh(mainwin);
			return -1;
		}
		else
		{

			aim_conn_addhandler(authconn, AIM_CB_FAM_GEN, AIM_CB_GEN_ERROR, faimtest_auth_error, 0);
			aim_conn_addhandler(authconn, AIM_CB_FAM_SPECIAL, AIM_CB_SPECIAL_AUTHSUCCESS, faimtest_auth_success, 0);
			aim_conn_addhandler(authconn, AIM_CB_FAM_GEN, AIM_CB_GEN_SERVERREADY, faimtest_authsvrready, 0);
			aim_send_login(authconn, getUserID().toCString(), getPassword().toCString(), &info);
		}
		
		while(getRunStatus()==crsRun)
		{
			if (aim_queue_outgoing) aim_tx_flushqueue();
			
			if (aim_get_command() < 0)
			{
				printf("\a");
				wprintw(mainwin, "BONIM: connection error!\n\r");
				wrefresh(mainwin);
			}
			else aim_rxdispatch();
			
			// now we check for keypresses
			
			/* Watch stdin (fd 0) to see when it has input. */
			FD_ZERO(&rfds);
			FD_SET(0, &rfds);
			/* Wait up to 200ms. */
			tv.tv_sec = 0;
			tv.tv_usec = 200000;
			retval = select(1, &rfds, NULL, NULL, &tv);
			
			if (retval) process_input();
			
			wrefresh(inputwin);
		}
		
		/* Close up */
		wprintw(mainwin, "Writing config file.\n\r");
		wrefresh(mainwin);
		writeConfig();
		wprintw(mainwin, "Logging off AOL...\n\r");
		wrefresh(mainwin);
		/* close up all connections, dead or no */
		aim_logoff();
	}
	
	wprintw(mainwin, "This has been a bonehead production.\n\r");
	wrefresh(mainwin);
	
	/* Get out */
	endwin();
	return 0;
}

Buddy *Client::getBuddyPtrByID(String budID)
{
	unsigned int iter;
	Buddy *iterBud;
	
	budID=unformatBudID(budID);
	
	for(iter=0, iterBud=m_buds.getItem(0); iter<m_buds.getSize(); iter++, iterBud=m_buds.getItem(iter))
	{
		if(!iterBud) return NULL;
		
		if(budID==unformatBudID(iterBud->budID)) return iterBud;
	}
	
	return NULL;
}

String Client::unformatBudID(String budID)
{
	String returnMe;
	unsigned long iter;
	
	budID=budID.toLower().toUpper();
	
	for(iter=0; iter<budID.getTokenCount(); iter++) returnMe=returnMe+budID.getToken(iter);
	
	return returnMe;
}

int faimtest_serverready(struct command_rx_struct *command, ...)
{
	switch (command->conn->type)
	{
		case AIM_CONN_TYPE_BOS:
		aim_bos_reqrate(command->conn); /* request rate info */
		aim_bos_ackrateresp(command->conn);  /* ack rate info response -- can we say timing? */
		aim_bos_setprivacyflags(command->conn, 0x00000003);
		
		#if 0
		aim_bos_reqpersonalinfo(command->conn);
		#endif
		
		aim_bos_reqservice(command->conn, AIM_CONN_TYPE_ADS); /* 0x05 == Advertisments */
		
		#if 0
		aim_bos_reqrights(NULL);
		aim_bos_reqbuddyrights(NULL);
		aim_bos_reqlocaterights(NULL);
		aim_bos_reqicbmparaminfo(NULL);
		#endif
		
		/* set group permissions */
		aim_bos_setgroupperm(NULL, 0x1f);
		wprintw(BONIM->mainwin, "BONIM: done with BOS ServerReady\n\r");
		wrefresh(BONIM->mainwin);
		break;
		case AIM_CONN_TYPE_CHATNAV:
		wprintw(BONIM->mainwin, "BONIM: chatnav: got server ready\n\r");
		wrefresh(BONIM->mainwin);
		break;
		default:
		wprintw(BONIM->mainwin, "BONIM: unknown connection type on Server Ready\n\r");
		wrefresh(BONIM->mainwin);
	}
	return 1;
}

/*
handleredirect()...

This, of course, handles Service Redirects from OSCAR.

Should get passed in the following:
struct command_rx_struct *command
the raw command data
int serviceid
the destination service ID
char *serverip
the IP address of the service's server
char *cookie
the raw auth cookie
*/

int faimtest_handleredirect(struct command_rx_struct *command, ...)
{
	va_list ap;
	int serviceid;
	char *ip;
	char *cookie;
	
	
	/* this is the new buddy list */
	char buddies[5120];
	
	strcpy(buddies, BONIM->AOLUserList.toCString());
	va_start(ap, command);
	serviceid = va_arg(ap, int);
	ip = va_arg(ap, char *);
	cookie = va_arg(ap, char *);
	va_end(ap);
	
	switch(serviceid)
	{
		case 0x0005: /* Advertisements */
		{
			/*
			* The craziest explanation yet as to why we finish logging in when
			* we get the advertisements redirect, of which we don't use anyway....
			*                    IT WAS EASY!
			*/
			
			/* send the buddy list and profile (required, even if empty) */
			aim_bos_setbuddylist(command->conn, buddies);
			if(BONIM->getVisStatus()==visBuds) BONIM->makeBudsOnlyVis(command->conn);
			refreshprofile();
			
			/* send final login command (required) */
			aim_bos_clientready(command->conn); /* tell BOS we're ready to go live */
			
			/* you should now be ready to go */
			wprintw(BONIM->mainwin, "\n\rYou are now officially online. :) (%s)\n\r\n\r", ip);
			wrefresh(BONIM->mainwin);
			
			if(BONIM->m_hitAndRunTarget!="")
			{
				sleep(1);	/* Give AOL time to notify people who have you as a buddy so that it does not IM before your target sees you as on-line. If you IM before someone thinks you are on-line and he has "knock-knocks" on then they get a "knock-knock" dialog even when you are on their buddy list. */
				
				aim_send_im(NULL, BONIM->m_hitAndRunTarget.toCString(), 0, BONIM->m_hitAndRunMessage.toCString());
				
				quitBonimInt(0);
			}
			
			break;
		}
		case 0x0007: /* Authorizer */
		{
			struct aim_conn_t *tstconn;
			/* Open a connection to the Auth */
			tstconn = aim_newconn(AIM_CONN_TYPE_AUTH, ip);
			if ( (tstconn==NULL) || (tstconn->status >= AIM_CONN_STATUS_RESOLVERR) )
			wprintw(BONIM->mainwin, "BONIM: unable to reconnect with authorizer\n\r");
			else
			/* Send the cookie to the Auth */
			aim_auth_sendcookie(tstconn, cookie);
			
			wrefresh(BONIM->mainwin);
			
			break;
		}
		case 0x000d: /* ChatNav */
		{
			struct aim_conn_t *tstconn = NULL;
			tstconn = aim_newconn(AIM_CONN_TYPE_CHATNAV, ip);
			if ( (tstconn==NULL) || (tstconn->status >= AIM_CONN_STATUS_RESOLVERR))
			{
				wprintw(BONIM->mainwin, "BONIM: unable to connect to chatnav server\n\r");
				wrefresh(BONIM->mainwin);
				return 1;
			}
			aim_conn_addhandler(tstconn, AIM_CB_FAM_CTN, AIM_CB_SPECIAL_DEFAULT, aim_parse_unknown, 0);
			aim_conn_addhandler(tstconn, AIM_CB_FAM_GEN, AIM_CB_SPECIAL_DEFAULT, aim_parse_unknown, 0);
			aim_auth_sendcookie(tstconn, cookie);
			printf("\a");
			wprintw(BONIM->mainwin, "chatnav: connected\n\r");
			wrefresh(BONIM->mainwin);
		}
		break;
		case 0x000e: /* Chat */
		{
			#if 0
			struct aim_conn_t *tstconn = NULL;
			tstconn = aim_newconn(AIM_CONN_TYPE_CHAT, ip);
			if ( (tstconn==NULL) || (tstconn->status >= AIM_CONN_STATUS_RESOLVERR))
			{
				wprintw(BONIM->mainwin, "BONIM: unable to connect to chat server\n\r");
				wrefresh(BONIM->mainwin);
				return 1;
			}
			aim_auth_sendcookie(aim_getconn_type(AIM_CONN_TYPE_CHAT), cookie);
			printf("\a");
			wprintw(BONIM->mainwin, "chat: connected\n\r");
			wrefresh(BONIM->mainwin);
			#endif
			
			break;
		}
		default:
		wprintw(BONIM->mainwin,"uh oh... got redirect for unknown service 0x%04x!!\n\r", serviceid);
		wrefresh(BONIM->mainwin);
		/* dunno */
	}
	
	return 1;
}

int faimtest_auth_error(struct command_rx_struct *command, ...)
{
	va_list ap;
	struct login_phase1_struct *logininfo;
	char *errorurl;
	short errorcode;
	
	va_start(ap, command);
	logininfo = va_arg(ap, struct login_phase1_struct *);
	wprintw(BONIM->mainwin,"Screen name: %s\n\r", logininfo->screen_name);
	errorurl = va_arg(ap, char *);
	wprintw(BONIM->mainwin,"Error URL: %s\n\r", errorurl);
	errorcode = va_arg(ap, int);	/* Compiler fix: was "short" */
	wprintw(BONIM->mainwin,"Error code: 0x%02x\n\r", errorcode);
	wprintw(BONIM->mainwin,"If you got an error code of 0x00 and a url of\n\rhttp://www.aol.com you're probably connecting\n\rtoo fast, wait 10 minutes and try again.\n\r");
	va_end(ap);
	wrefresh(BONIM->mainwin);
	
	aim_conn_close(aim_getconn_type(AIM_CONN_TYPE_AUTH));
	
	endwin();
	
	quitBonimInt(0);
	
	return 0;
}

int faimtest_auth_success(struct command_rx_struct *command, ...)
{
	va_list ap;
	struct login_phase1_struct *logininfo;
	struct aim_conn_t *bosconn = NULL;
	
	va_start(ap, command);
	logininfo = va_arg(ap, struct login_phase1_struct *);
	va_end(ap);
	
	wprintw(BONIM->mainwin, "Screen name: %s\n\r", logininfo->screen_name);
	//  wprintw(mainwin,"Reg status: %2d\n\r", logininfo->regstatus);
	wprintw(BONIM->mainwin, "Email: %s\n\r", logininfo->email);
	//  wprintw(mainwin,"Cookie len: %d\n\r", sizeof(logininfo->cookie));
	//  wprintw(mainwin,"BOS IP: %s\n\r", logininfo->BOSIP);
	
	wprintw(BONIM->BONIM->mainwin, "Closing auth connection...\n\r");
	wrefresh(BONIM->mainwin);
	
	aim_conn_close(command->conn);
	bosconn = aim_newconn(AIM_CONN_TYPE_BOS, logininfo->BOSIP);
	if (bosconn == NULL) wprintw(BONIM->mainwin, "BONIM: could not connect to BOS: internal error\n\r");
	else if (bosconn->status != 0) wprintw(BONIM->mainwin, "BONIM: could not connect to BOS\n\r");
	else
	{
		aim_conn_addhandler(bosconn, AIM_CB_FAM_ACK, AIM_CB_ACK_ACK, NULL, 0);
		aim_conn_addhandler(bosconn, AIM_CB_FAM_GEN, AIM_CB_GEN_SERVERREADY, faimtest_serverready, 0);
		aim_conn_addhandler(bosconn, AIM_CB_FAM_GEN, AIM_CB_GEN_RATEINFO, NULL, 0);
		aim_conn_addhandler(bosconn, AIM_CB_FAM_GEN, AIM_CB_GEN_REDIRECT, faimtest_handleredirect, 0);
		aim_conn_addhandler(bosconn, AIM_CB_FAM_GEN, AIM_CB_GEN_MOTD, NULL, 0);
		aim_conn_addhandler(bosconn, AIM_CB_FAM_STS, AIM_CB_STS_SETREPORTINTERVAL, NULL, 0);
		aim_conn_addhandler(bosconn, AIM_CB_FAM_BUD, AIM_CB_BUD_ONCOMING, faimtest_parse_oncoming, 0);
		aim_conn_addhandler(bosconn, AIM_CB_FAM_BUD, AIM_CB_BUD_OFFGOING, faimtest_parse_offgoing, 0);
		aim_conn_addhandler(bosconn, AIM_CB_FAM_MSG, AIM_CB_MSG_INCOMING, faimtest_parse_incoming_im, 0);
		aim_conn_addhandler(bosconn, AIM_CB_FAM_LOC, AIM_CB_LOC_ERROR, faimtest_parse_misses, 0);
		aim_conn_addhandler(bosconn, AIM_CB_FAM_MSG, AIM_CB_MSG_MISSEDCALL, faimtest_parse_misses, 0);
		aim_conn_addhandler(bosconn, AIM_CB_FAM_GEN, AIM_CB_GEN_RATECHANGE, faimtest_parse_misses, 0);
		aim_conn_addhandler(bosconn, AIM_CB_FAM_MSG, AIM_CB_MSG_ERROR, faimtest_parse_misses, 0);
		aim_conn_addhandler(bosconn, AIM_CB_FAM_LOC, AIM_CB_LOC_USERINFO, faimtest_parse_userinfo, 0);
		
		aim_conn_addhandler(bosconn, AIM_CB_FAM_CTN, AIM_CB_CTN_DEFAULT, aim_parse_unknown, 0);
		aim_conn_addhandler(bosconn, AIM_CB_FAM_SPECIAL, AIM_CB_SPECIAL_DEFAULT, aim_parse_unknown, 0);
		aim_auth_sendcookie(bosconn, logininfo->cookie);
	}
	return 1;
	wrefresh(BONIM->mainwin);
}

int faimtest_parse_userinfo(struct command_rx_struct *command, ...)
{
	struct aim_userinfo_s *userinfo;
	char *prof_encoding = NULL;
	char *prof = NULL;
	
	va_list ap;
	va_start(ap, command);
	userinfo = va_arg(ap, struct aim_userinfo_s *);
	prof_encoding = va_arg(ap, char *);
	prof = va_arg(ap, char *);
	va_end(ap);
	
	wprintw(BONIM->mainwin, "BONIM: userinfo: sn: %s\n\r", userinfo->sn);
	wprintw(BONIM->mainwin, "BONIM: userinfo: warnlevel: 0x%04x\n\r", userinfo->warnlevel);
	wprintw(BONIM->mainwin, "BONIM: userinfo: class: 0x%04x = ", userinfo->uclass);
	wrefresh(BONIM->mainwin);
	
	/*
	*  00000000  (binary)
	*         1  Trial
	*        2   Unknown
	*       3    AOL
	*      4     Unknown
	*     5      Free
	*
	* ORed together.
	*
	*/
	
	if (userinfo->uclass & 0x0001)
	wprintw(BONIM->mainwin,"TRIAL ");
	if (userinfo->uclass & 0x0002)
	wprintw(BONIM->mainwin,"UNKNOWN_BIT2 ");
	if (userinfo->uclass & 0x0004)
	wprintw(BONIM->mainwin,"AOL ");
	if (userinfo->uclass & 0x0008)
	wprintw(BONIM->mainwin,"UNKNOWN_BIT4 ");
	if (userinfo->uclass & 0x0010)
	wprintw(BONIM->mainwin,"FREE ");
	wprintw(BONIM->mainwin,"\n\r");
	
	wprintw(BONIM->mainwin,"BONIM: userinfo: membersince: %lu\n\r", userinfo->membersince);
	wprintw(BONIM->mainwin,"BONIM: userinfo: onlinesince: %lu\n\r", userinfo->onlinesince);
	wprintw(BONIM->mainwin,"BONIM: userinfo: idletime: 0x%04x\n\r", userinfo->idletime);
	
	wprintw(BONIM->mainwin,"BONIM: userinfo: profile_encoding: %s\n\r", prof_encoding ? prof_encoding : "[none]");
	wprintw(BONIM->mainwin,"BONIM: userinfo: prof: %s\n\r", prof ? prof : "[none]");
	wrefresh(BONIM->mainwin);
	
	return 1;
}

/*
* The user-level Incoming ICBM callback.
*
* Arguments:
*  struct command_rx_struct *  command     if you feel like doing it yourself
*  char *                      srcsn       the source name
*  char *                      msg         message
*  int                         warnlevel   warning/evil level
*  int                         class       user class
*  ulong                       membersince time_t of date of signup
*  ulong                       onsince     time_t of date of singon
*  int                         idletime    min (sec?) idle
*  int                         isautoreply TRUE if its an auto-response
*
*/

int faimtest_parse_incoming_im(struct command_rx_struct *command, ...)
{
	char loglocation[256];
	struct aim_userinfo_s *userinfo;
	char *msg = NULL;
	int isautoreply = 0, ppos, pflag, cval;
	int mpos;
	va_list ap;
	//  char *tmpstr = NULL;
	char uclass[4] = "---";
	char fnick[20] = "---";
	char tbuffer[4096];	
	ofstream fOut;
	
	va_start(ap, command);
	userinfo = va_arg(ap, struct aim_userinfo_s *);
	msg = va_arg(ap, char *);
	isautoreply = va_arg(ap, int);
	va_end(ap);
	
	sprintf(fnick,"%s",userinfo->sn);
	if (userinfo->uclass & 0x0010)
	sprintf(uclass,"[F]");
	else if (userinfo->uclass & 0x0001)
	sprintf(uclass,"[T]");
	else if (userinfo->uclass & 0x0004)
	sprintf(uclass,"[A]");
	else
	sprintf(uclass,"[?]");
	
	BONIM->setLastFrom(fnick);
	BONIM->displayStatus();
	
	if(BONIM->beepOnMsg()) cout << "\a";
	
	if(BONIM->logIMs())
	{
		strcpy(loglocation, BONIM->getConfigDir().toCString());
		strcat(loglocation, "logs/");
		strcat(loglocation, fnick);
		
		fOut.open(loglocation, ios::app | ios::out);
	}
	
	wattron(BONIM->mainwin, COLOR_PAIR(1));
	wprintw(BONIM->mainwin, "<%s> ",fnick);
	wattron(BONIM->mainwin, COLOR_PAIR(5));
	
	if(BONIM->logIMs()) fOut << '<' << fnick << '>';
	
	// This section is just to strip HTML tags, I know it could be better, but
	// it only has to work :P - I've not been helped by the fact AIM send < and
	// > in the messages AS < & > rather than &lt; and &gt; which is what I'd
	// expect - I can only assume this is to allow AIM users to include HTML
	// in their messages (which seems daft to me) Oh well...
	
	ppos=0;
	pflag=0;
	for (mpos=0; mpos<=(int)strlen(msg); mpos++)
	{
		if (msg[mpos]=='<')
		pflag=1;
		if (msg[mpos]=='>')
		pflag=0;
		if (pflag==0 && msg[mpos]!='>')
		{
			tbuffer[ppos]=msg[mpos];
			if (tbuffer[ppos]==13)
			tbuffer[ppos]='*';
			cval=tbuffer[ppos];
			if (cval>=128)
			tbuffer[ppos]='?';
			ppos++;
		}
	}
	tbuffer[ppos]='\0';
	
	wprintw(BONIM->mainwin, "%s\n\r",tbuffer);
	wattron(BONIM->mainwin, COLOR_PAIR(6));
	wrefresh(BONIM->mainwin);
	
	if(BONIM->logIMs()) fOut << tbuffer << "\n\r";
	
	if(BONIM->getAwayMsg()!="")
	{
		aim_send_im(NULL, fnick, AIM_IMFLAGS_AWAY, BONIM->getAwayMsg().toCString());
		
		wattron(BONIM->mainwin, COLOR_PAIR(7));
		wprintw(BONIM->mainwin, "-> autoresponse: ");
		wattron(BONIM->mainwin, COLOR_PAIR(6));
		wprintw(BONIM->mainwin, "%s\n\r", BONIM->getAwayMsg().toCString());
		wattron(BONIM->mainwin, COLOR_PAIR(7));
		wrefresh(BONIM->mainwin);
		
		if(BONIM->logIMs())
		{
			strcpy(loglocation, BONIM->getConfigDir().toCString());
			strcat(loglocation, "logs/");
			strcat(loglocation, BONIM->getLastFrom().toCString());
			
			fOut << "-> autoresponse: " << BONIM->getAwayMsg() << "\r\n";
		}
	}
	
	if(BONIM->logIMs()) fOut.close();
	
	return 1;
}

int faimtest_authsvrready(struct command_rx_struct *command, ...)
{
	wprintw(BONIM->mainwin,"faimtest_authsvrready: called (contype: %d)\n\r", command->conn->type);
	sleep(10);
	/* should just be able to tell it we're ready too... */
	aim_auth_clientready(command->conn);
	
	#if 0
	/*
	* This is where you'd really begin changing your password.
	*   However, this callback may get called for reasons other
	*   than you wanting to change your password.  You should
	*   probably check that before actually doing it.
	*/
	aim_auth_changepasswd(command->conn, "PWD1", "PWD2");
	#endif
	wrefresh(BONIM->mainwin);
	
	return 1;
}

int faimtest_pwdchngdone(struct command_rx_struct *command, ...)
{
	wprintw(BONIM->mainwin,"PASSWORD CHANGE SUCCESSFUL!!!\n\r");
	wrefresh(BONIM->mainwin);
	return 1;
}

int faimtest_parse_oncoming(struct command_rx_struct *command, ...)
{
	struct aim_userinfo_s *userinfo;
	char tnick[32];
	Buddy bud;
	String timeStr;
	
	va_list ap;
	va_start(ap, command);
	userinfo = va_arg(ap, struct aim_userinfo_s *);
	va_end(ap);
	
	wrefresh(BONIM->mainwin);
	sprintf(tnick, "%s", userinfo->sn);
	
	BONIM->htimeval = time(0);
	BONIM->thetime=localtime(&BONIM->htimeval);
	
	bud=BONIM->getBudByID(tnick);
	
	if(bud.budStatus==avsOffLine && BONIM->showSigns())
	{
		wattron(BONIM->mainwin, COLOR_PAIR(2));
		wprintw(BONIM->mainwin, "* Logon  @ %.02d:%.02d %s (%s)\n\r", BONIM->thetime->tm_hour, BONIM->thetime->tm_min, tnick, (char *)bud.budAlias.toCString());
		wattron(BONIM->mainwin, COLOR_PAIR(6));
	}
	
	wrefresh(BONIM->mainwin);
	
	if (userinfo->uclass & 0x0010) BONIM->setBudClass(bud.budID, uclFree);	//sprintf(tbuffer,"[F]");
	else if (userinfo->uclass & 0x0001) BONIM->setBudClass(bud.budID, uclTrial);	//sprintf(tbuffer,"[T]");
	else if (userinfo->uclass & 0x0004) BONIM->setBudClass(bud.budID, uclAOL);	//sprintf(tbuffer,"[A]");
	else BONIM->setBudClass(bud.budID, uclUnknown);	//sprintf(tbuffer,"[?]");
	
	BONIM->setBudStatus(bud.budID, avsOnLine);
	
	if(BONIM->thetime->tm_hour<10) timeStr=String(' ')+String(long(BONIM->thetime->tm_hour))+':';
	else timeStr=String(long(BONIM->thetime->tm_hour))+':';
	
	if(BONIM->thetime->tm_min<10) timeStr=timeStr+'0'+String(long(BONIM->thetime->tm_min));
	else timeStr=timeStr+String(long(BONIM->thetime->tm_min));
	
	BONIM->setBudTime(bud.budID, timeStr);
	
	BONIM->showUserList();
	return 1;
}

int faimtest_parse_offgoing(struct command_rx_struct *command, ...)
{
	char tnick[32];
	Buddy bud;
	String timeStr;
	
	sprintf(tnick,"%s",&(command->data[11]));
	
	BONIM->htimeval = time(0);
	BONIM->thetime=localtime(&BONIM->htimeval);
	
	bud=BONIM->getBudByID(tnick);
	
	if(BONIM->showSigns())
	{
		wattron(BONIM->mainwin, COLOR_PAIR(1));
		wprintw(BONIM->mainwin, "* Logoff @ %.02d:%.02d %s (%s)\n\r", BONIM->thetime->tm_hour,BONIM->thetime->tm_min,tnick,bud.budAlias.toCString());
		wattron(BONIM->mainwin, COLOR_PAIR(6));
		wrefresh(BONIM->mainwin);
	}
	
	BONIM->setBudStatus(bud.budID, avsOffLine);
	
	if(BONIM->thetime->tm_hour<10) timeStr=String(' ')+String(long(BONIM->thetime->tm_hour))+':';
	else timeStr=String(long(BONIM->thetime->tm_hour))+':';
	
	if(BONIM->thetime->tm_min<10) timeStr=timeStr+'0'+String(long(BONIM->thetime->tm_min));
	else timeStr=timeStr+String(long(BONIM->thetime->tm_min));
	
	BONIM->setBudTime(bud.budID, timeStr);
	
	BONIM->showUserList();
	return 1;
}


/*
* Handles callbacks for: AIM_CB_RATECHANGE, AIM_CB_USERERROR,
*   AIM_CB_MISSED_IM, and AIM_CB_MISSED_CALL.
*/

int faimtest_parse_misses(struct command_rx_struct *command, ...)
{
	u_short family;
	u_short subtype;
	
	family = (command->data[0] << 8) + command->data[1];
	subtype = (command->data[2] << 8) + command->data[3];
	
	/*
	switch (family)
	{
		case 0x0001:
		if (subtype == 0x000a)
		wprintw(mainwin,"*** You are exceeding the speed limit! slow down, heh.\n\r");
		break;
		case 0x0002:
		if (subtype == 0x0001)
		{
			u_long snacid = 0x00000000;
			
			snacid = aimutil_get32(&command->data[6]);
			
			wprintw(mainwin,"Received unknown error in SNAC family 0x0002 (snacid = %08lx)\n\r", snacid);
		}
		break;
		case 0x0004:
		if (subtype == 0x0001)
		wprintw(mainwin,"*** Message bounced - requested user not online\n\r");
		else if (subtype == 0x000a)
		wprintw(mainwin,"You missed some messages from %s because they were sent too fast\n\r", &(command->data[13]));
		break;
	}
	wrefresh(mainwin);
	*/
	
	return 0;
}

void refreshprofile()
{
	ifstream fIn;
	char proflocation[256];
	char profile[4097];
	
	strcpy(proflocation, BONIM->getConfigDir().toCString());
	strcat(proflocation, "profile");
	
	fIn.open(proflocation);
	
	if(!fIn)
	{
		ofstream fOut;
		
		fOut.open(proflocation);
		
		strcpy(profile, "Yo! I'm using Bonim - an AIM client for the linux console. You can get the latest source code at freshmeat.net or you can email David at theexternvoid@yahoo.com for more info.");
		fOut << profile;
		
		fOut.close();
	}
	else
	{
		fIn.read(profile, 4096);
		
		fIn.close();
	}
	
	aim_bos_setprofile(NULL, profile);
}

void process_input(void)
{
	char tc;
	int ct;
	char tbuffer[4096];	
	
	tc=getch();
	ct=tc;
	
	//  *** This is for debugging, will dump some data to the screen on keypress.
	//
	//  sprintf(tbuffer,"*DEBUG* C:%c A:%d S:%s P:%d\n\r",tc, ct, BONIM->getInputLine(), BONIM->getInputLine().getLength());
	//  waddstr(mainwin,tbuffer);
	//  wrefresh(mainwin);
	
	if (ct != 10 && ct != 7 && ct != 8 && ct != 127 && ct != 21 && ct > 31 && ct < 128)
	{
		BONIM->appendInputLine(tc);
		sprintf(tbuffer, "%c", tc);
		waddstr(BONIM->inputwin, tbuffer);
		wrefresh(BONIM->inputwin);
	}
	else
	{
		if (ct == 10)
		{
			waddstr(BONIM->inputwin, "\n\r");
			process_command(BONIM->getInputLine());
			BONIM->clearInputLine();
			if(BONIM->getRunStatus()==crsRun) BONIM->showPrompt();	//Why is this tested with an if statement?
		}
		if (ct == 21)
		{
			waddstr(BONIM->inputwin, " *FLUSHED*\n\r");
			BONIM->showPrompt();
		}
		if ((ct == 7 || ct == 8 || ct == 127) && BONIM->getInputLine()!="")
		{
			BONIM->chopInputLine();
			waddstr(BONIM->inputwin,"\b \b");
			wrefresh(BONIM->inputwin);
		}
	}
}

void display_help_cmdprompt(void)
{
	cout << "Here is a list of all commands and their shortcuts:" << endl;
	cout << endl;
	cout << "/ADD nick   --  Adds nick to buddy list." << endl;
	cout << "/AWAY       /A  Sets an away message. It uses the" << endl;
	cout << "                EDITOR environment variable to open" << endl;
	cout << "                your favorite editor to edit the" << endl;
	cout << "                away message. You can specify a name for" << endl;
	cout << "                an away message: \"/AWAY <name>\".  This will" << endl;
	cout << "                load a saved away message or create a new" << endl;
	cout << "                one if it does not exist. \"/AWAY LIST\" will" << endl;
	cout << "                list the saved away messages. You currently" << endl;
	cout << "                have to delete them manually. They are" << endl;
	cout << "                saved in \"~/.bonim/<your IM name>/away/\"." << endl;
	cout << "                Note that if you set an away message with" << endl;
	cout << "                no name then it is stored as \"default\"" << endl;
	cout << "                automatically, so naming is optional." << endl;
	cout << "/BEEP       /B  Toggles beeping on messages." << endl;
	cout << "/CLEAR      /C  Clears the chat and input windows." << endl;
	cout << "/DEL nick   --  Deletes nick from buddy list." << endl;
	cout << "/HELP       /H  List basic commands." << endl;
	cout << "/LOG        /L  Toggles IM logging. Logs are stored in" << endl;
	cout << "                ~/.bonim/<profilename>/logs/." << endl;
	cout << "/LOGOFF     --  Same as \"/QUIT\"." << endl;
	cout << "/OFFLBUDS   /O  Toggles off-line names in buddy list." << endl;
	cout << "/PROFILE    --  Let's you set your user profile. It uses" << endl;
	cout << "                the EDITOR environment variable to open" << endl;
	cout << "                your favorite editor to edit the profile." << endl;
	cout << "/QUERY usr  /Q  Selects target user for messages." << endl;
	cout << "                Use screen name or aliases. It also" << endl;
	cout << "                selects users not on your buddy list." << endl;
	cout << "                If number, select that user on list." << endl;
	cout << "                So \"/q 3\" will select the 3rd" << endl;
	cout << "                on-line buddy. To select nobody, leave" << endl;
	cout << "                usr blank." << endl;
	cout << "/QUIT       --  Terminate client and log off AIM." << endl;
	cout << "/REPLY msg  /R  Sends 'msg' to the last user to IM you." << endl;
	cout << "                with no msg it selects them as target." << endl;
	cout << "/RESTART    --  Restart (logout then log back in)." << endl;
	cout << "/SALIAS txt --  Set alias of current user to txt." << endl;
	cout << "/SHOWSIGNS  --  Toggle showing sign on/off messages." << endl;
	cout << "/STATUS     --  See info on your status such as your" << endl;
	cout << "                current away message, if logging is on," << endl;
	cout << "                etc." << endl;
	cout << "/TALIAS     --  Swap between screen names and aliases." << endl;
	cout << "/TLIST      /T  Switch 80 col. userlist (status/time)." << endl;
	cout << "/VIS        --  Sets your visibility. Do \"/vis all\" to" << endl;
	cout << "                be visible to all, \"vis buds\" to be" << endl;
	cout << "                visible only to your buddy list. In the" << endl;
	cout << "                latter case, you will be shown as off-line" << endl;
	cout << "                to non-buddies and you will not be able to IM" << endl;
	cout << "                non-buddies." << endl;
	cout << "/WHO        /W  Show userlist in short format." << endl;
	cout << endl;
	cout << endl;
	
	errUsageGlobal();
}

void send_reply(const char *cmd)
{
	char tempmessage[4096];
	char tbuffer[4096];
	
	if ((strcmp(cmd,"/r") == 0)|| (strcmp(cmd,"/reply") == 0) || (strcmp(cmd,"/r ") == 0) || (strcmp(cmd,"/reply ") == 0))
	{
		BONIM->setTarget(BONIM->getLastFrom());
		sprintf(tbuffer,"Target '%s' selected.\n\r", BONIM->getTarget().toCString());
		waddstr(BONIM->mainwin, tbuffer);
		wrefresh(BONIM->mainwin);
	}
	else
	{
		if (cmd[2] == ' ')
		{
			strcpy(tempmessage,&cmd[3]);
		}
		else
		{
			strcpy(tempmessage,&cmd[7]);
		}
		
		if(BONIM->getLastFrom()=="--- None ---")
		{
			waddstr(BONIM->mainwin, "BONIM: No one to reply to.\n\r");
			wrefresh(BONIM->mainwin);
			return;
		}
		
		BONIM->unaway();
		aim_send_im(NULL, BONIM->getLastFrom().toCString(), 0, tempmessage);
		wattron(BONIM->mainwin, COLOR_PAIR(7));
		wprintw(BONIM->mainwin,"-> %s: ",BONIM->getLastFrom().toCString());
		wattron(BONIM->mainwin, COLOR_PAIR(6));
		wprintw(BONIM->mainwin, "%s\n\r",tempmessage);
		wattron(BONIM->mainwin, COLOR_PAIR(7));
		wrefresh(BONIM->mainwin);
		
		if(BONIM->logIMs())
		{
			ofstream fOut;
			
			fOut.open((BONIM->getConfigDir()+"logs/"+BONIM->getLastFrom()).toCString(), ios::app | ios::out);
			fOut << "-> " << BONIM->getLastFrom().toCString() << ": " << tempmessage << "\r\n";
			fOut.close();
		}
		
		BONIM->setLastTo(BONIM->getLastFrom());
		BONIM->displayStatus();
	}
}

void process_command(const String &cmd)
{
	char loglocation[256];
	String tok0;
	char tbuffer[4096];
		
	if(!cmd.getLength()) return;
	
	if(cmd[0] == '/')
	{
		tok0=cmd.getToken(0).mid(1).toLower();	//Get the first token, chop of the left '/' character, and make it lower case ala the Swiss Army String class.  :)
		
		if(tok0=="tlist" || tok0=="t")
		{
			if(BONIM->getLowColDisp()==lcdStatus)
			{
				BONIM->setLowColDisp(lcdTime);
				wattron(BONIM->mainwin, COLOR_PAIR(1));
				wprintw(BONIM->mainwin, "Userlist will now show logon/logoff times.\n\r");
				wrefresh(BONIM->mainwin);
				wattron(BONIM->mainwin, COLOR_PAIR(6));
				BONIM->showUserList();
			}
			else
			{
				BONIM->setLowColDisp(lcdStatus);
				wattron(BONIM->mainwin, COLOR_PAIR(1));
				wprintw(BONIM->mainwin, "Userlist will now show status information.\n\r");
				wrefresh(BONIM->mainwin);
				wattron(BONIM->mainwin, COLOR_PAIR(6));
				BONIM->showUserList();
			}
		}
		else if(tok0=="away" || tok0=='a')
		{
			String awayFile;
			
			/* In case it's missing, make the away directory... */
			mkdir((BONIM->getConfigDir()+"away").toCString(), 0755);
			
			/* Yes, I know that this next line is uber cheesy... */
			if(cmd.getTokens(1, 0).toLower()=="list") system((String("clear; ls '")+BONIM->getConfigDir()+"away/' | less").toCString());
			else
			{
				if(BONIM->getAwayMsg()!="") BONIM->unaway();
				else
				{
					ifstream fIn;
					
					awayFile=cmd.getTokens(1, 0);
					
					if(awayFile=="") awayFile=BONIM->getConfigDir()+"away/default";
					else awayFile=BONIM->getConfigDir()+"away/"+awayFile;
					
					if(!getenv("EDITOR"))
					{
						wprintw(BONIM->mainwin, "You need to set your ENVIRONMENT system variable.\n\rRead the README file for details.\n\r");
						
						return;
					}
					
					system((String(getenv("EDITOR"))+" '"+awayFile+"'").toCString());
					
					BONIM->clearAll();
					
					fIn.open(awayFile.toCString());
					
					if(fIn)
					{
						String inputLine;
						
						BONIM->setAwayMsg(inputLine.readRaw(fIn, 4096));
						
						wattron(BONIM->mainwin, COLOR_PAIR(1));
						wprintw(BONIM->mainwin, "Away status turned on:\n\r");
						wattron(BONIM->mainwin, COLOR_PAIR(5));
						wprintw(BONIM->mainwin, "%s\n\r", BONIM->getAwayMsg().toCString());
						wrefresh(BONIM->mainwin);
						wattron(BONIM->mainwin, COLOR_PAIR(6));
						
						fIn.close();
					}
					else
					{
						wattron(BONIM->mainwin, COLOR_PAIR(1));
						wprintw(BONIM->mainwin,"Error: away message file could not be read.\n\r");
						wrefresh(BONIM->mainwin);
						wattron(BONIM->mainwin, COLOR_PAIR(6));
					}
				}
			}
			
			BONIM->displayStatus();
		}
		else if(tok0=="profile")
		{
			if(!getenv("EDITOR"))
			{
				wprintw(BONIM->mainwin, "You need to set your ENVIRONMENT system variable.\n\rRead the README file for details.\n\r");
				
				return;
			}
			
			system((String(getenv("EDITOR"))+" \""+BONIM->getConfigDir()+"profile\"").toCString());
			
			BONIM->clearAll();
			
			BONIM->displayStatus();
			
			refreshprofile();
			
			wattron(BONIM->mainwin, COLOR_PAIR(1));
			wprintw(BONIM->mainwin, "Your profile has been set.\n\r");
			wrefresh(BONIM->mainwin);
			wattron(BONIM->mainwin, COLOR_PAIR(6));
		}
		else if(tok0=="offlbuds" || tok0=='o')
		{
			if(BONIM->showOffline())
			{
				BONIM->setShowOffline(false);
				wattron(BONIM->mainwin, COLOR_PAIR(1));
				wprintw(BONIM->mainwin, "Userlist will not show off-line buddies.\n\r");
				wrefresh(BONIM->mainwin);
				wattron(BONIM->mainwin, COLOR_PAIR(6));
				BONIM->showUserList();
			}
			else
			{
				BONIM->setShowOffline(true);
				wattron(BONIM->mainwin, COLOR_PAIR(1));
				wprintw(BONIM->mainwin,"Userlist will show off-line buddies.\n\r");
				wrefresh(BONIM->mainwin);
				wattron(BONIM->mainwin, COLOR_PAIR(6));
				BONIM->showUserList();
			}
		}
		else if(tok0=="log" || tok0=='l')
		{
			if(BONIM->logIMs())
			{
				BONIM->setLogIMs(false);
				wattron(BONIM->mainwin, COLOR_PAIR(1));
				wprintw(BONIM->mainwin,"IM logging is now off.\n\r");
				wrefresh(BONIM->mainwin);
				wattron(BONIM->mainwin, COLOR_PAIR(6));
				BONIM->showUserList();
			}
			else
			{
				BONIM->setLogIMs(true);
				wattron(BONIM->mainwin, COLOR_PAIR(1));
				wprintw(BONIM->mainwin,"IM logging is now on.\n\r");
				wrefresh(BONIM->mainwin);
				wattron(BONIM->mainwin, COLOR_PAIR(6));
				BONIM->showUserList();
			}
		}
		else if(tok0=="log" || tok0=='l')
		{
			if(BONIM->logIMs())
			{
				BONIM->setLogIMs(false);
				wattron(BONIM->mainwin, COLOR_PAIR(1));
				wprintw(BONIM->mainwin,"IM logging is now off.\n\r");
				wrefresh(BONIM->mainwin);
				wattron(BONIM->mainwin, COLOR_PAIR(6));
				BONIM->showUserList();
			}
			else
			{
				BONIM->setLogIMs(true);
				wattron(BONIM->mainwin, COLOR_PAIR(1));
				wprintw(BONIM->mainwin,"IM logging is now on.\n\r");
				wrefresh(BONIM->mainwin);
				wattron(BONIM->mainwin, COLOR_PAIR(6));
				BONIM->showUserList();
			}
		}
		else if(tok0=="showsigns")
		{
			if(BONIM->showSigns())
			{
				BONIM->setShowSigns(false);
				wattron(BONIM->mainwin, COLOR_PAIR(1));
				wprintw(BONIM->mainwin, "Sign on/off messages are disabled.\n\r");
				wrefresh(BONIM->mainwin);
				wattron(BONIM->mainwin, COLOR_PAIR(6));
				BONIM->showUserList();
			}
			else
			{
				BONIM->setShowSigns(true);
				wattron(BONIM->mainwin, COLOR_PAIR(1));
				wprintw(BONIM->mainwin, "Sign on/off messages are enabled.\n\r");
				wrefresh(BONIM->mainwin);
				wattron(BONIM->mainwin, COLOR_PAIR(6));
				BONIM->showUserList();
			}
		}
		else if(tok0=="status")
		{
			wattron(BONIM->mainwin, COLOR_PAIR(1));
			
			if(BONIM->beepOnMsg()) wprintw(BONIM->mainwin, "Beeping on incoming messages is enabled.\n\r");
			else wprintw(BONIM->mainwin, "Beeping on incoming messages is disabled.\n\r");
			
			if(BONIM->logIMs()) wprintw(BONIM->mainwin, "Logging is enabled.\n\r");
			else wprintw(BONIM->mainwin, "Logging is disabled.\n\r");
			
			if(BONIM->showOffline()) wprintw(BONIM->mainwin, "Offline buddies are shown.\n\r");
			else wprintw(BONIM->mainwin, "Offline buddies aren't shown.\n\r");
			
			if(BONIM->getVisStatus()==visBuds) wprintw(BONIM->mainwin, "You are invisible to non-buddies.\n\r");
			else wprintw(BONIM->mainwin, "You are visible to all.\n\r");
			
			if(BONIM->showSigns()==visBuds) wprintw(BONIM->mainwin, "Sign on/off messages are enabled.\n\r");
			else wprintw(BONIM->mainwin, "Sign on/off messages are disabled.\n\r");
			
			if(BONIM->getAwayMsg()!="")
			{
				wprintw(BONIM->mainwin, "You are away with this message:\n\r");
				wattron(BONIM->mainwin, COLOR_PAIR(5));
				wprintw(BONIM->mainwin, (char *)BONIM->getAwayMsg().toCString());
			}
			else wprintw(BONIM->mainwin, "You are not away.\n\r");
			
			wrefresh(BONIM->mainwin);
			wattron(BONIM->mainwin, COLOR_PAIR(6));
		}
		else if(tok0=="beep" || tok0=='b')
		{
			if(BONIM->beepOnMsg())
			{
				BONIM->setBeepOnMsg(false);
				wattron(BONIM->mainwin, COLOR_PAIR(1));
				wprintw(BONIM->mainwin,"Beeping on incoming messages is now disabled.\n\r");
				wrefresh(BONIM->mainwin);
				wattron(BONIM->mainwin, COLOR_PAIR(6));
			}
			else
			{
				BONIM->setBeepOnMsg(true);
				wattron(BONIM->mainwin, COLOR_PAIR(1));
				wprintw(BONIM->mainwin, "Beeping on incoming messages is now enabled.\n\r");
				wrefresh(BONIM->mainwin);
				wattron(BONIM->mainwin, COLOR_PAIR(6));
			}
		}
		else if(tok0=="vis")
		{
			String argRest=cmd.getTokens(1, 0).toLower();
			
			if(argRest!="")
			{
				wattron(BONIM->mainwin, COLOR_PAIR(1));
				
				if(argRest=="all")
				{
					BONIM->setVisStatus(visAll);
					
					wprintw(BONIM->mainwin, "You are completely visible.\n\r");
				}
				else if(argRest=="buds")
				{
					wprintw(BONIM->mainwin, "All people not on your buddy list are banished.\n\r");
					
					BONIM->setVisStatus(visBuds);
				}
				else wprintw(BONIM->mainwin, "Invalid parameter. Enter \"all\" or \"buds\".\n\r");
				
				wattron(BONIM->mainwin, COLOR_PAIR(6));
				wrefresh(BONIM->mainwin);
			}
			else
			{
				sprintf(tbuffer,"You must give an argument to this command.\n\r");
				waddstr(BONIM->mainwin, tbuffer);
				wrefresh(BONIM->mainwin);
			}
		}
		else if(tok0=="salias")
		{
			String argRest=cmd.getTokens(1, 0);
			
			if(argRest=="")
			{
				sprintf(tbuffer,"You must give an argument to this command.\n\r");
				waddstr(BONIM->mainwin, tbuffer);
				wrefresh(BONIM->mainwin);
			}
			else
			{
				String newAlias;
				
				newAlias=cmd.mid(8);
				sprintf(tbuffer, "Setting alias...\n\r");
				waddstr(BONIM->mainwin, tbuffer);
				
				if(!BONIM->setBudAlias(BONIM->getTarget(), newAlias)) BONIM->showUserList();
				else wprintw(BONIM->mainwin, "Unable to add alias, %s is not in buddy list.\n\r", BONIM->getTarget().toCString());
				
				wrefresh(BONIM->mainwin);
			}
		}
		else if(tok0=="query" || tok0=="q")
		{
			String argRest=cmd.getTokens(1, 0);
			
			if(argRest=="")
			{
				BONIM->setTarget("<none>");
				sprintf(tbuffer, "Unselected target.\n\r");
				waddstr(BONIM->mainwin, tbuffer);
				wrefresh(BONIM->mainwin);
			}
			else
			{
				strcpy(cTempCrud, argRest.toCString());
				BONIM->getBuddy(cTempCrud, cTempCrud, 0, 1);
				BONIM->setTarget(cTempCrud);
				
				sprintf(tbuffer, "Target '%s' selected.\n\r", BONIM->getTarget().toCString());
				waddstr(BONIM->mainwin, tbuffer);
				wrefresh(BONIM->mainwin);
			}
		}
		else if(tok0=="help" || tok0=='h') BONIM->displayHelp();
		else if(tok0=="logout" || tok0=="quit") BONIM->setRunStatus(crsQuit);
		else if(tok0=="who" || tok0=="w") BONIM->showNames();
		else if(tok0=="add") BONIM->addUser();
		else if(tok0=="reply" || tok0=='r') send_reply(cmd.toCString());
		else if(tok0=="del") BONIM->delUser();
		else if(tok0=="restart") BONIM->setRunStatus(crsRestart);
		else if(tok0=="clear" || tok0=='c') BONIM->clearAll();
		else
		{
			waddstr(BONIM->mainwin, "Invalid command.\n\r");
			wrefresh(BONIM->mainwin);
		}
	}
	else
	{
		if(BONIM->getTarget()=="<none>")
		{
			waddstr(BONIM->mainwin,"BONIM: No target selected.\n\r");
			wrefresh(BONIM->mainwin);
			return;
		}
		
		BONIM->setLastTo(BONIM->getTarget());
		BONIM->displayStatus();
		
		if(BONIM->getAwayMsg()!="") BONIM->unaway();
		
		wattron(BONIM->mainwin, COLOR_PAIR(7));
		wprintw(BONIM->mainwin, "[%s] ", BONIM->getUserID().toCString());
		wattron(BONIM->mainwin, COLOR_PAIR(6));
		wprintw(BONIM->mainwin, "%s\n\r", BONIM->getInputLine().toCString());
		wattron(BONIM->mainwin, COLOR_PAIR(7));
		wrefresh(BONIM->mainwin);
		
		if(BONIM->logIMs())
		{
			ofstream fOut;
			
			strcpy(loglocation, BONIM->getConfigDir().toCString());
			strcat(loglocation, "logs/");
			strcat(loglocation, BONIM->getTarget().toCString());
			
			fOut.open(loglocation, ios::app | ios::out);
			fOut << '[' << BONIM->getUserID() << "] " << BONIM->getInputLine().toCString() << "\r\n";
			fOut.close();
		}
		
		aim_send_im(NULL, BONIM->getTarget().toCString(), 0, cmd.toCString());
	}
}
