#ifndef PTRLIST_TEMPLATE
#define PTRLIST_TEMPLATE

#include <string.h>

#include "ptrlistnode.h"

template <class PTRLISTTYPE>
class PtrList
{
private:
	PtrListNode<PTRLISTTYPE> *head;
	mutable PtrListNode<PTRLISTTYPE> *lastAccessed;
	unsigned long size;
	mutable unsigned long lastPos;
	mutable bool lastPosDirty;
	
public:
	PtrList()
	//PURPOSE-Default constructor
	//POST-The list is constructed with default values
	{
		head=NULL;
		size=0;
		
		lastPosDirty=true;
	}
	
	virtual ~PtrList()
	//PURPOSE-Destructor
	//POST-The list is destructed
	{
		kill();
	}
	
	void kill()
	//PURPOSE-To destroy the list's contents
	//POST-The list's contents are destroyed
	{
		while(size!=0) killItem(0);
	}

	bool isEmpty() const
	//PURPOSE-Is the list empty?
	//OUT-Whether or not the list is empty
	//POST-If the list is empty, true is returned; otherwise false is returned
	{
		return size==0;
	}

	unsigned long getSize() const
	//PURPOSE-To get the list's length
	//OUT-The list's length
	//POST-The list's length is returned
	{
		return size;
	}
	
	bool addItem(PTRLISTTYPE *newItem, unsigned long position)
	//PURPOSE-To insert an item into the list
	//IN-newItem, position
	//OUT-Whether or not the addition was successful
	//POST-If the item can be inserted, it is and true is returned; else false is returned
	{
		PtrListNode<PTRLISTTYPE> *newNode;
		PtrListNode<PTRLISTTYPE> *tempNode;
		unsigned long iter;
		
		if(position<0 || position>size) return false;
		
		newNode=new PtrListNode<PTRLISTTYPE>(newItem);
		
		if(!newNode) return false;
		
		if(head==NULL) head=newNode;
		else if(position==size)
		{
			for(tempNode=head; tempNode->getNext()!=NULL; tempNode=tempNode->getNext());
			
			tempNode->setNext(newNode);
			newNode->setPrevious(tempNode);
		}
		else
		{
			for(iter=0, tempNode=head; iter<position; iter++, tempNode=tempNode->getNext());
			
			newNode->setNext(tempNode);
			newNode->setPrevious(tempNode->getPrevious());
			tempNode->setPrevious(newNode);
			
			tempNode=newNode->getPrevious();
			
			if(tempNode!=NULL) tempNode->setNext(newNode);
			else head=newNode;
		}
		
		size++;
		if(position<=lastPos) lastPos++;
		
		return true;
	}
	
	bool killItem(unsigned long position)
	//PURPOSE-To delete the item at the passed position
	//IN-position
	//OUT-Whether or not the deletion was successful
	//POST-If the item at position can be deleted, it is and true is returned; else false is returned
	{
		PtrListNode<PTRLISTTYPE> *tempNode;
		unsigned long iter;
		
		if(position<0 || position>size) return false;
		
		for(iter=0, tempNode=head; iter<position; iter++, tempNode=tempNode->getNext());
		
		if(tempNode==head) head=head->getNext();

		delete tempNode;
		size--;
		
		if(position<lastPos) lastPos--;
		else if(position==lastPos) lastPosDirty=true;
		
		return true;
	}
	
	PTRLISTTYPE *releaseItem(unsigned long position)
	//PURPOSE-To get the item at the passed position and remove it from the list
	//IN-position
	//OUT-The item at the passed position
	//POST-If the item can be retrieved it is returned, else NULL is returned; whe nit is retrieved it is reomved from the list but NOT deallocated like killItem() does
	{
		PtrListNode<PTRLISTTYPE> *tempNode;
		PTRLISTTYPE *tempData;
		unsigned long iter;
		
		if(position<0 || position>size-1) return 0;
		
		if(((position>lastPos)?position-lastPos:lastPos-position)>position || lastPosDirty)
		{
			for(iter=0, tempNode=head; iter!=position; iter++, tempNode=tempNode->getNext())
			{
			}
			
			lastAccessed=tempNode;
			lastPos=position;
			lastPosDirty=false;
		}
		else
		{
			if(position<lastPos)
			{
				for(; lastPos!=position; lastPos--, lastAccessed=lastAccessed->getPrevious());
			}
			else
			{
				for(; lastPos!=position; lastPos++, lastAccessed=lastAccessed->getNext());
			}
			
			tempNode=lastAccessed;
		}
		
		tempData=tempNode->releaseData();
		
		killItem(position);
		
		return tempData;
	}
	
	PTRLISTTYPE *getItem(unsigned long position) const
	//PURPOSE-To get the item at the passed position
	//IN-position
	//OUT-The item at the passed position
	//POST-If the item can be retrieved it is returned, else NULL is returned
	{
		PtrListNode<PTRLISTTYPE> *tempNode;
		unsigned long iter;
		
		if(position<0 || position>size-1 || !size) return 0;
		
		if(((position>lastPos)?position-lastPos:lastPos-position)>position || lastPosDirty)
		{
			for(iter=0, tempNode=head; iter!=position; iter++, tempNode=tempNode->getNext());
			
			lastAccessed=tempNode;
			lastPos=position;
			lastPosDirty=false;
		}
		else
		{
			if(position<lastPos)
			{
				for(; lastPos!=position; lastPos--, lastAccessed=lastAccessed->getPrevious());
			}
			else
			{
				for(; lastPos!=position; lastPos++, lastAccessed=lastAccessed->getNext());
			}
			
			tempNode=lastAccessed;
		}
		
		return tempNode->getData();		
	}
};

#endif
